% Wavelet packet ICA example script on human faces (2D) 

% External packages needed:
% FastICA
% BSICA available at http://appsrv.cse.cuhk.edu.hk/~kzhang/research.html

% MATLAB Toolboxes needed:
% Signal Processing Toolbox
% Wavelet Toolbox

%%%%%%%%%%%%%%%%%%%%%
% USER FLAGS
%%%%%%%%%%%%%%%%%%%%%

% set to 1 to enable, or to 0 disable the feature
addWhiteGaussianNoiseAsDependentComponent = 1;
nonDecimatedWaveletPacketTransform = 0;

% number of levels of the wavelet packet decomposition
%   warning: if the non decimated decomposition is chosen, 
%   more then 2 levels may be computationally very demanding
level = 2;

% choose the nonlinearity of the FastICA ('pow3', 'tanh', 'gaus')
nonlinearity = 'tanh';

%%%%%%%%%%%%%%%%%%%%%
% end of USER FLAGS
%%%%%%%%%%%%%%%%%%%%%

% set the MATLAB path
wd = pwd; p = strcat(wd, '\wp'); path(path, p);
 
% read the images (four human faces)
ims1 = im2double(rgb2gray(imread('data/x43.jpg'))); ims2 = im2double(rgb2gray(imread('data/x45.jpg')));
ims3 = im2double(rgb2gray(imread('data/x46.jpg'))); ims4 = im2double(rgb2gray(imread('data/x50.jpg')));
[r, c] = size(ims1); datalen = r*c;

% form one-dimensional vectors from images
s1 = snake(ims1); s2 = snake(ims2); s3 = snake(ims3); s4 = snake(ims4); 

% subtract DC component
s1 = s1 - mean(s1);      s2 = s2 - mean(s2);      s3 = s3 - mean(s3);     s4 = s4 - mean(s4);

if addWhiteGaussianNoiseAsDependentComponent
    s5 = 0.008*randn(1,length(s1)); % approximately 29 dB S/N ratio
    s1 = s1 + s5; s2 = s2 + s5; s3 = s3 + s5; s4 = s4 + s5;
end

% form images from one-dimensional vectors 
ims1 = isnake(s1, r, c); ims2 = isnake(s2, r, c); ims3 = isnake(s3, r, c); ims4 = isnake(s4, r, c);
  
% show input images
h = figure;
subplot(5,4,1), myimshow(ims1, []), subplot(5,4,2), myimshow(ims2, []);
subplot(5,4,3), myimshow(ims3, []), subplot(5,4,4), myimshow(ims4, []);
subplot(5,4, 1), title('A) Source images')
disp('A) Source images. Press enter to continue...')
pause

% form mixed signals ("observed" data)
S = [s1; s2; s3; s4];
% A = randn(4,4); save data/A A % random mixing matrix
load data/A
X = A * S;

% extract the mixed signals
x1 = X(1,:); x2 = X(2,:); x3 = X(3,:); x4 = X(4,:); 

% form images from the 1D mixtures
imx1 = isnake(x1, r, c); imx2 = isnake(x2, r, c); imx3 = isnake(x3, r, c); imx4 = isnake(x4, r, c);

% show the images
figure(h),
subplot(5,4,5), myimshow(imx1, []), subplot(5,4,6), myimshow(imx2, []);
subplot(5,4,7), myimshow(imx3, []), subplot(5,4,8), myimshow(imx4, []);
subplot(5,4, 5), title('B) Observed images')
disp('B) Observed images. Press enter to continue...')
pause

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% direct application of the ICA method
[Aa, W] = fastica(X,'numOfIC', 4, 'approach', 'symm', 'g', nonlinearity, 'displayMode', 'off');

% compute recovered signals
Y = W*X;

ae_direct = amari_error(W*A)

% extract the recovered signals
y1 = Y(1,:); y2 = Y(2,:); y3 = Y(3,:); y4 = Y(4,:); 

% form images from the 1D signals
ys1 = isnake(y1, r,c); ys2 = isnake(y2, r,c); ys3 = isnake(y3, r,c); ys4 = isnake(y4, r,c);

% show the images
figure(h),
subplot(5,4, 9), myimshow(ys1, []), subplot(5,4,10), myimshow(ys2, []);
subplot(5,4,11), myimshow(ys3, []), subplot(5,4,12), myimshow(ys4, []);
subplot(5,4, 9), title('C) Directly recovered images')
disp('C) Directly recovered images. Press enter to continue...')
pause

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% innovation approach
t0=clock;
a1=lpc(x1,10); a2=lpc(x2,10); a3=lpc(x3,10); a4=lpc(x4,10);
a=(a1+a2+a3+a4)/4;
x1i=filter(a,1,x1); x2i=filter(a,1,x2); 
x3i=filter(a,1,x3); x4i=filter(a,1,x4); 

[Aainn, Winn] = fastica([x1i; x2i; x3i; x4i],'numOfIC', 4, 'approach', 'symm', 'g', nonlinearity, 'displayMode', 'off');

% compute recovered signals
Yinv = Winn*X;
time_inn=etime(clock,t0)

ae_innovations = amari_error(Winn*A)

% form images from the 1D signals
yin1 = isnake(Yinv(1,:), r,c); yin2 = isnake(Yinv(2,:), r,c); yin3 = isnake(Yinv(3,:), r,c); yin4 = isnake(Yinv(4,:), r,c);

% show the images
h1 = figure;
subplot(2,4,1), myimshow(yin1, []), subplot(2,4,2), myimshow(yin2, []);
subplot(2,4,3), myimshow(yin3, []), subplot(2,4,4), myimshow(yin4, []);
subplot(2,4,1), title('Innovation approach')
disp('Innovation approach has ended. Press enter to continue...')
pause

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% LP-HP filter approach (the first level WP transform)
t0=clock;

swp1 = swpdec(x1, 1, 'sym8'); swp2 = swpdec(x2, 1, 'sym8');
swp3 = swpdec(x3, 1, 'sym8'); swp4 = swpdec(x4, 1, 'sym8');
a1 = swp1(2,:); a2 = swp2(2,:); a3 = swp3(2,:); a4 = swp4(2,:);     % take HP part as being more independent
[Ahp, Whp] = fastica([a1(:), a2(:), a3(:), a4(:)]','numOfIC', 4, 'approach', 'symm', 'g', nonlinearity, 'displayMode', 'off');

% compute recovered signals
Yhp = Whp*X;
time_hp=etime(clock,t0)

ae_HPfiltered = amari_error(Whp*A)

% form images from the 1D signals
yhp1 = isnake(Yhp(1,:), r,c); yhp2 = isnake(Yhp(2,:), r,c); yhp3 = isnake(Yhp(3,:), r,c); yhp4 = isnake(Yhp(4,:), r,c);

% show the images
figure(h1)
subplot(2,4,5), myimshow(yhp1, []), subplot(2,4,6), myimshow(yhp2, []);
subplot(2,4,7), myimshow(yhp3, []), subplot(2,4,8), myimshow(yhp4, []);
subplot(2,4,5), title('HP filtering approach')
disp('HP filtering approach has ended. Press enter to continue.')
pause
disp('Starting WPSDICA approach...')

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% 2D wavelet packet (WP) approach
t0=clock;
if nonDecimatedWaveletPacketTransform
    [swp1,lp] = swpdec2(imx1, level, 'sym8'); swp2 = swpdec2(imx2, level, 'sym8');
        swp3  = swpdec2(imx3, level, 'sym8'); swp4 = swpdec2(imx4, level, 'sym8');
else
    [swp1,lp] = dwpdec2(imx1, level, 'sym8'); swp2 = dwpdec2(imx2, level, 'sym8');
        swp3  = dwpdec2(imx3, level, 'sym8'); swp4 = dwpdec2(imx4, level, 'sym8');
end

% compute approximation of the mutual information (MI)
lsum = size(lp,1);
MI = zeros(lsum+1, 1);

MI(1) = MI_measure(X); % MI of the input mixture

% MI of the wavelet packet subbands
for i = 1:lsum,
    if nonDecimatedWaveletPacketTransform
        a1 = swp1(:,:,i); a2 = swp2(:,:,i); a3 = swp3(:,:,i); a4 = swp4(:,:,i);
    else
        a1 = dwpcoef2(swp1,i); a2 = dwpcoef2(swp2,i); a3 = dwpcoef2(swp3,i); a4 = dwpcoef2(swp4,i);
    end
    MI(i+1) = MI_measure([a1(:), a2(:), a3(:), a4(:)]'); 
end

% normalization
MI = MI / max(MI);

% find the WP subband of the minimum mutal information
[MImni, mind] = min(MI(2:end));
 
% WP subband ICA
if nonDecimatedWaveletPacketTransform
    a1 = swp1(:,:,mind); a2 = swp2(:,:,mind); a3 = swp3(:,:,mind); a4 = swp4(:,:,mind);
else
    a1 = dwpcoef2(swp1,mind); a2 = dwpcoef2(swp2,mind); a3 = dwpcoef2(swp3,mind); a4 = dwpcoef2(swp4,mind);
end
[Aawp, Wwp] = fastica([a1(:), a2(:), a3(:), a4(:)]','numOfIC', 4, 'approach', 'symm', 'g', nonlinearity, 'displayMode', 'off');

% compute recovered signals
Ywp = Wwp*X;
time_wp=etime(clock,t0)

ae_wp = amari_error(Wwp*A)

% show the MI tree
showMI2(MI)

% extract the recovered signals
ywp1 = Ywp(1,:); ywp2 = Ywp(2,:); ywp3 = Ywp(3,:); ywp4 = Ywp(4,:);

% form images from the 1D signals
ywps1 = isnake(ywp1, r,c); ywps2 = isnake(ywp2, r,c); ywps3 = isnake(ywp3, r,c); ywps4 = isnake(ywp4, r,c);

% show the recovered images
figure(h),
subplot(5,4,13), myimshow(ywps1, []), subplot(5,4,14), myimshow(ywps2, []);
subplot(5,4,15), myimshow(ywps3, []), subplot(5,4,16), myimshow(ywps4, []);
subplot(5,4,13), title('D) WPSDICA approach');

disp('D) WPSDICA approach has ended. Press enter to continue.')
pause
disp('Starting Zhan-Chan BSICA approach...')

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Zhang - Chan algorithm
[yI, y, hh, Wzc] = bsica(X, 11);

ae_Chan = amari_error(Wzc*A)

% extract the recovered signals
yaf1 = y(1,:); yaf2 = y(2,:); yaf3 = y(3,:); yaf4 = y(4,:);

% form images from the 1D signals
yf1 = isnake(yaf1, r,c); yf2 = isnake(yaf2, r,c); yf3 = isnake(yaf3, r,c); yf4 = isnake(yaf4, r,c);

% show the recovered images
figure(h),
subplot(5,4,17), myimshow(yf1, []), subplot(5,4,18), myimshow(yf2, []);
subplot(5,4,19), myimshow(yf3, []), subplot(5,4,20), myimshow(yf4, []);
subplot(5,4,17), title('E) Zhang-Chan approach')

disp('E) Zhang-Chan approach has ended.');
